import pandas as pd
from typing import List
from abc import ABC, abstractmethod


class FalsificationAlgorithm(ABC):
    def __init__(self) -> None:
        """
        Base class for all falsification algorithms.
        """
        super().__init__()

    @abstractmethod
    def test(
        self,
        data: pd.DataFrame,
        covariate_vars: List[str],
        treatment_var: str,
        outcome_var: str,
        source_var: str,
    ) -> float:
        """
        Run the falsification test to test null hypothesis of no unmeasured confounding in multi-source observational data.

        Args:
            data (pd.DataFrame): Input dataset.
            covariate_vars (List[str]): List of covariate column names.
            treatment_var (str): Treatment variable column name.
            outcome_var (str): Outcome variable column name.
            source_var (str): Source/environment indicator column_
        
        Returns:
            float: p-value from the test.
        """
        pass


from typing import Optional
import numpy as np
from causallearn.utils.cit import CIT

"""
Multiple implementations of conditional independence tests to test null hypothesis:

    H0: X ⟂ Y | Z

"""


def _validate_inputs(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> None:
    """
    Validate input shapes for CI tests.
    """
    if not (x.ndim == 2 and x.shape[1] == 1):
        raise ValueError(
            f"Input x must be a 2D array with a single column (shape: (n_samples, 1)), but has shape {x.shape}."
        )
    if not (y.ndim == 2 and y.shape[1] == 1):
        raise ValueError(
            f"Input y must be a 2D array with a single column (shape: (n_samples, 1)), but has shape {y.shape}."
        )
    if not (z.ndim == 2):
        raise ValueError(
            f"Input z must be a 2D array with shape (n_samples, n_features), but has shape {z.shape}."
        )

    n_samples = x.shape[0]
    if not (y.shape[0] == n_samples == z.shape[0]):
        raise ValueError("All inputs must have the same number of rows.")


def kcit_rbf(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> Optional[float]:
    """
    Kernel-based Conditional Independence Test (KCIT) with RBF kernels.

    Args:
        x: Array of shape (n_samples, 1).
        y: Array of shape (n_samples, 1).
        z: Array of shape (n_samples, n_covariates).

    Returns:
        p-value of the test if successful; None if an error occurred.
    """
    _validate_inputs(x, y, z)

    data = np.hstack([x, y, z])

    try:
        cit_obj = CIT(
            data,
            method="kci",
            kernelX="Gaussian",
            kernelY="Gaussian",
            kernelZ="Gaussian",
            approx=False,
            use_gp=True,
            est_width="median",
        )
        conditioning_set = list(range(2, data.shape[1]))
        pval = cit_obj(0, 1, conditioning_set)
        assert 0 <= pval <= 1, f"Invalid p-value computed: {pval}"
        return pval

    except AssertionError as e:
        print(f"Assertion error: {e}")
        return None


def fisherz(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> Optional[float]:
    """
    Fisher's Z Conditional Independence Test.

    Args:
        x: Array of shape (n_samples, 1).
        y: Array of shape (n_samples, 1).
        z: Array of shape (n_samples, n_covariates).

    Returns:
        p-value of the test if successful; None if an error occurred.
    """
    _validate_inputs(x, y, z)

    data = np.hstack([x, y, z])

    try:
        cit_obj = CIT(data, method="fisherz")
        conditioning_set = list(range(2, data.shape[1]))
        pval = cit_obj(0, 1, conditioning_set)
        assert 0 <= pval <= 1, f"Invalid p-value computed: {pval}"
        return pval

    except AssertionError as e:
        print(f"Assertion error: {e}")
        return None



from causal_falsify.abstract import FalsificationAlgorithm
import numpy as np
import pandas as pd
from typing import List

from causal_falsify.utils.cond_indep import (
    kcit_rbf,
    fisherz,
)


class TransportabilityTest(FalsificationAlgorithm):
    def __init__(
        self,
        cond_indep_test: str = "kcit_rbf",
        max_sample_size: int = np.inf,
    ) -> None:
        """

        Transportability-based test akin to the benchmarking framework in
        "Using Trial and Observational Data to Assess Effectiveness:
        Trial Emulation, Transportability, Benchmarking, and Joint Analysis"
        Dahabreh et al., 2024

        Joint test for whether we have transportability and unconfoundedness across sources.
        A rejection will falsify both conditions jointly.

        Args:
            cond_indep_test (str): CI test to use: "kcit_rbf" or "fisherz".
            max_sample_size (int): Max samples for testing to control runtime.
        """
        super().__init__()
        if max_sample_size <= 0:
            raise ValueError("max_sample_size must be larger than zero")

        self.cond_indep_test = cond_indep_test
        self.max_sample_size_test = max_sample_size

    def test(
        self,
        data: pd.DataFrame,
        covariate_vars: List[str],
        treatment_var: str,
        outcome_var: str,
        source_var: str,
    ) -> float:
        """
        Perform falsification test for joint test of unconfoundedness and transportability.

        Args:
            data (pd.DataFrame): DataFrame containing all required columns.
            covariate_vars (List[str]): Covariate column names to condition on.
            treatment_var (str): Treatment column name.
            outcome_var (str): Outcome column name.
            source_var (str): Source/environment indicator column name.

        Returns:
            float: p-value of the conditional independence test.
        """
        # Validate required columns
        required_cols = set(covariate_vars + [treatment_var, outcome_var, source_var])
        missing = required_cols.difference(data.columns)
        if missing:
            raise ValueError(f"Missing columns in data: {missing}")

        # Extract arrays for the test
        outcome = data[[outcome_var]].values  # shape (n_samples, 1)
        treatment = data[[treatment_var]].values  # shape (n_samples, 1)
        source = data[[source_var]].values  # shape (n_samples, 1)
        covariates = data[covariate_vars].values  # shape (n_samples, n_covariates)

        # Subsample if necessary
        if outcome.shape[0] > self.max_sample_size_test:
            outcome, source, covariates, treatment = self.subsample_data(
                outcome, source, covariates, treatment
            )

        # Select conditional independence test function
        if self.cond_indep_test == "kcit_rbf":
            test_func = kcit_rbf
        elif self.cond_indep_test == "fisherz":
            test_func = fisherz
        else:
            raise ValueError(f"Unsupported cond_indep_test: {self.cond_indep_test}")

        # Test if outcome is independent of source conditional on covariates and treatment
        conditioning_vars = np.hstack([covariates, treatment])
        pval = test_func(outcome, source, conditioning_vars)

        return pval

    def subsample_data(self, outcome, source, covariates, treatment):
        """
        Subsample data to limit size, preserving source distribution proportions.
        """
        unique_sources, counts = np.unique(source, return_counts=True)
        proportions = counts / counts.sum()

        sampled_indices = []
        for src_value, proportion in zip(unique_sources, proportions):
            src_indices = np.where(source.flatten() == src_value)[0]
            n_samples = min(
                len(src_indices), int(np.round(proportion * self.max_sample_size_test))
            )
            sampled_indices.extend(
                np.random.choice(src_indices, n_samples, replace=False)
            )

        if len(sampled_indices) > self.max_sample_size_test:
            sampled_indices = np.random.choice(
                sampled_indices, self.max_sample_size_test, replace=False
            )

        return (
            outcome[sampled_indices, :],
            source[sampled_indices, :],
            covariates[sampled_indices, :],
            treatment[sampled_indices, :],
        )



import numpy as np
import jax.numpy as jnp
from jax.scipy.linalg import solve
from jax import grad, jit, random, lax
from sklearn.preprocessing import PolynomialFeatures


def create_polynomial_representation(
    X, degree, use_sklearn: bool = False, interaction_only: bool = False
):
    if degree <= 1:
        raise ValueError("Degree must be larger than 2.")

    if interaction_only and not use_sklearn:
        print("Warning: interaction_only has no effect as use_sklearn = False.")

    if use_sklearn:
        return PolynomialFeatures(
            degree=degree, interaction_only=interaction_only, include_bias=False
        ).fit_transform(X)

    else:
        n_features = X.shape[1]

        # Create an empty list to store polynomial features
        poly_features = []

        # Iterate over each feature
        for feature_idx in range(n_features):
            # Create polynomial features for the current feature
            feature = X[:, feature_idx]
            poly_feature = np.column_stack([feature**d for d in range(1, degree + 1)])
            poly_features.append(poly_feature)

        # Stack the polynomial features horizontally
        X_poly = np.hstack(poly_features)
        return X_poly


###############################################################
# Test based on computing Frobenius norm of off-diagonal block
###############################################################


def compute_offdiag_block_frobnorm(data_x, data_y):

    dim_x, dim_y = data_x.shape[1], data_y.shape[1]
    assert data_x.shape[0] == data_y.shape[0], "first dimension be the same"
    coefs = np.hstack([data_x, data_y])

    validate_matrix(coefs)

    covariance_matrix = np.cov(coefs, rowvar=False)
    offdiag_block = covariance_matrix[:dim_x, dim_x:]
    assert offdiag_block.shape == (dim_x, dim_y)

    return np.linalg.norm(offdiag_block, "fro")


def permutation_independence_test(
    data_x: np.ndarray, data_y: np.ndarray, n_bootstraps: int = 1000, random_state=None
) -> float:

    if random_state is None:
        random_state = np.random.RandomState()

    observed_frob_norm = compute_offdiag_block_frobnorm(data_x, data_y)

    resampled_frob_norm = np.zeros((n_bootstraps, 1))
    for j in range(n_bootstraps):

        # permute rows in coef_t
        permuted_data_x = random_state.permutation(data_x)  # permutates on first axis
        resampled_frob_norm[j] = compute_offdiag_block_frobnorm(permuted_data_x, data_y)

    return np.mean(observed_frob_norm < resampled_frob_norm)


def bootstrapped_permutation_independence_test(
    data_x: np.ndarray,
    data_y: np.ndarray,
    resampled_data_x: np.ndarray,
    resampled_data_y: np.ndarray,
    random_state=None,
) -> float:
    if random_state is None:
        random_state = np.random.RandomState()

    n_bootstraps = resampled_data_x.shape[0]

    assert resampled_data_x.shape[:1] == resampled_data_y.shape[:1]

    observed_frob_norm = compute_offdiag_block_frobnorm(data_x, data_y)

    resampled_frob_norm = np.zeros((n_bootstraps, 1))
    for j in range(n_bootstraps):

        permuted_resampled_data_x = random_state.permutation(
            resampled_data_x[j, :, :].squeeze()
        )

        resampled_frob_norm[j] = compute_offdiag_block_frobnorm(
            permuted_resampled_data_x, resampled_data_y[j, :, :].squeeze()
        )

    return np.mean(observed_frob_norm < resampled_frob_norm)


##########################################
# Utils
##########################################


def validate_matrix(matrix: np.ndarray):
    # Assert that the input is a NumPy array
    assert isinstance(matrix, np.ndarray), "Input must be a NumPy array."

    # Assert no NaN values
    assert not jnp.isnan(matrix).any(), f"Matrix contains NaN values: {matrix}"

    # Assert no infinite values
    assert not np.isinf(matrix).any(), "Matrix contains infinite values."

    # Assert proper dimensionality
    assert matrix.ndim == 2, "Matrix must be 2-dimensional."


###############################################################
# Methods for estimating linear models
###############################################################


def fit_logistic_regression(X, T, alpha=1e-3):
    """
    Fit a logistic regression model using JAX.

    X: Transformed feature matrix (including intercept).
    T: Target variable (binary).
    alpha: Regularization parameter for ridge logistic regression.
    """

    # Define logistic regression loss with optional regularization
    def logistic_loss(params, X, T, alpha):
        logits = X @ params
        loss = jnp.mean(jnp.log(1 + jnp.exp(-T * logits)))  # Logistic loss
        if alpha > 0:
            loss += alpha * jnp.sum(params**2)  # Regularization term (Ridge)
        return loss

    # Initial guess for parameters (weights)
    init_params = jnp.zeros(X.shape[1])

    # Compute the gradient of the loss function
    loss_grad = grad(logistic_loss)

    # Minimize the loss using gradient descent
    def update(params, X, T, alpha, learning_rate=0.1):
        grads = loss_grad(params, X, T, alpha)
        return params - learning_rate * grads

    # Fit the model by iterating updates
    params = init_params
    for _ in range(1000):  # Set max iterations for gradient descent
        params = update(params, X, T, alpha)

    return params


def fit_linear_regression(X, Y, alpha=0):
    """
    Fit a linear regression model using JAX.

    X: Transformed feature matrix.
    Y: Target variable.
    alpha: Regularization parameter if using ridge regression (optional).
    """

    I = jnp.eye(X.shape[1])  # Identity matrix for regularization
    I = I.at[-1, -1].set(0)  # Exclude intercept from regularization
    params = solve(X.T @ X + alpha * I, X.T @ Y)
    return params


def cross_val_mse(X, Y, model_fn, num_folds):
    """
    Perform cross-validation and compute the mean squared error.

    X: Transformed feature matrix (including intercept).
    Y: Target variable.
    model_fn: Function to fit the model and return the parameters.
    num_folds: Number of folds for cross-validation.
    alpha: Regularization parameter (used for Ridge regression).
    """
    n = X.shape[0]
    fold_size = n // num_folds
    mse_list = []

    for i in range(num_folds):
        # Split data into training and validation sets
        val_indices = jnp.arange(i * fold_size, (i + 1) * fold_size)
        train_indices = jnp.concatenate(
            [jnp.arange(0, i * fold_size), jnp.arange((i + 1) * fold_size, n)]
        )

        X_train, X_val = X[train_indices], X[val_indices]
        Y_train, Y_val = Y[train_indices], Y[val_indices]

        # Fit the model and get parameters using the training set
        params = model_fn(X_train, Y_train)

        # Compute MSE on validation set
        preds = X_val @ params
        mse = jnp.mean((Y_val - preds) ** 2)
        mse_list.append(mse)

    return jnp.mean(jnp.array(mse_list))


def fit_outcome_model_jax(tf_XT, Y):

    assert tf_XT.shape[0] > tf_XT.shape[1], "need more samples than features"

    # Fit the outcome model using model_fn
    params_outcome = fit_linear_regression(tf_XT, Y)

    # Perform cross-validation for model diagnostic using the same model_fn
    model_mse = cross_val_mse(tf_XT, Y, fit_linear_regression, num_folds=5)

    return params_outcome.T, model_mse


def fit_treatment_model_jax(tf_X, T):

    assert tf_X.shape[0] > tf_X.shape[1], "need more samples than features"

    # Fit the model and get parameters
    params_treatment = fit_linear_regression(tf_X, T)

    # Perform cross-validation for model diagnostic using the same model_fn
    model_mse = cross_val_mse(tf_X, T, fit_linear_regression, num_folds=5)

    return params_treatment.T, model_mse


@jit
def bootstrap_model_fitting_jax(Y, T, tf_X, tf_XT, key):
    # Resample indices using JAX's random module for reproducibility
    key, subkey = random.split(key)  # Split the key to get a new one for resampling

    min_sample_size_needed_for_estimation = tf_X.shape[1] + 1
    assert (
        tf_X.shape[0] > min_sample_size_needed_for_estimation
    ), f"need more samples than {min_sample_size_needed_for_estimation}"
    resampled_indices = resample_until_enough_unique(
        subkey, Y.shape[0], min_sample_size_needed_for_estimation
    )

    # Resample the data
    resampled_Y = Y[resampled_indices]
    resampled_T = T[resampled_indices]
    resampled_tf_X = tf_X[resampled_indices]
    resampled_tf_XT = tf_XT[resampled_indices]

    # Fit outcome and treatment models on resampled data
    resampled_params_outcome, _ = fit_outcome_model_jax(resampled_tf_XT, resampled_Y)
    resampled_params_treatment, _ = fit_treatment_model_jax(resampled_tf_X, resampled_T)

    return resampled_params_outcome, resampled_params_treatment


def resample_until_enough_unique(subkey, n_resamples, min_sample_size):
    # Initial resampling
    resampled_indices = random.choice(
        subkey, n_resamples, shape=(n_resamples,), replace=True
    )

    def count_unique(x):
        x = jnp.sort(x)
        return 1 + (x[1:] != x[:-1]).sum()

    # Define condition function for while loop
    def condition_fn(state):
        _, resampled_indices = state
        # Check if unique indices are below the threshold
        return count_unique(resampled_indices) < min_sample_size

    # Define body function for while loop
    def body_fn(state):
        subkey, _ = state
        # Resample and update state
        subkey, new_subkey = random.split(subkey)
        resampled_indices = random.choice(
            new_subkey, n_resamples, shape=(n_resamples,), replace=True
        )
        return (subkey, resampled_indices)

    # Initial state: (key, resampled_indices)
    state = (subkey, resampled_indices)

    # Apply while loop until the condition is met
    _, resampled_indices = lax.while_loop(condition_fn, body_fn, state)

    return resampled_indices



import warnings
from typing import Dict, List, Any

import numpy as np
import pandas as pd
from scipy.stats import combine_pvalues
from sklearn.exceptions import ConvergenceWarning

from causal_falsify.abstract import FalsificationAlgorithm
from causal_falsify.utils.cond_indep import (
    kcit_rbf,
    fisherz,
)

warnings.simplefilter("ignore", ConvergenceWarning)


class HGIC(FalsificationAlgorithm):
    def __init__(
        self,
        cond_indep_test: str = "kcit_rbf",
        max_tests: int = -1,
        min_test_sample_size: int = 25,
        method_pval_combination: str = "tippett",
    ) -> None:
        """

        Hierarchical Graphical Independence Constraint (HGIC) algorithm from
        "Detecting Hidden Confounding in Observational Data Using Multiple Environments"
        Karlsson and Krijthe, NeurIPS 2023
        (https://arxiv.org/abs/2205.13935)

        Joint test for whether we have independence between causal mechanisms and unconfoundedness across sources.
        A rejection will falsify both conditions jointly.

        Args:
            cond_indep_test (str): CI test to use: "kcit_rbf" or "fisherz".
            max_tests (int): Maximum number of pairwise tests to perform.
            min_test_sample_size (int): Minimum sources per test.
            method_pval_combination (str): Method to combine p-values ("tippett", "fisher" or "stouffer").
        """
        super().__init__()

        if max_tests == 0:
            raise ValueError("max_tests must be non-zero (use -1 for unlimited).")

        self.cond_indep_test = cond_indep_test
        self.max_tests = max_tests
        self.min_test_sample_size = min_test_sample_size
        self.method_pval_combination = method_pval_combination
        # self.feature_scaling = self.independence_test_args.pop("feature_scaling", False)

    def test(
        self,
        data: pd.DataFrame,
        covariate_vars: List[str],
        treatment_var: str,
        outcome_var: str,
        source_var: str,
    ) -> float:
        """
        Perform falsification test for joint test of unconfoundedness and independence of causal mechanisms.

        Args:
            data (pd.DataFrame): DataFrame containing all required columns.
            covariate_vars (List[str]): Covariate column names to condition on.
            treatment_var (str): Treatment column name.
            outcome_var (str): Outcome column name.
            source_var (str): Source/environment indicator column name.

        Returns:
            float: p-value of the conditional independence test.
        """
        # Validate required columns
        required_cols = set(covariate_vars + [treatment_var, outcome_var, source_var])
        missing = required_cols.difference(data.columns)
        if missing:
            raise ValueError(f"Missing columns in data: {missing}")

        # Partition data by source variable
        grouped_data = {
            str(source_label): df for source_label, df in data.groupby(source_var)
        }

        # Sort sources by sample size
        sample_sizes = sorted(
            [(source_label, len(df)) for source_label, df in grouped_data.items()],
            key=lambda x: x[1],
            reverse=True,
        )
        if len(sample_sizes) < 2:
            raise ValueError("At least two sources are required for the test.")

        max_samples_available = sample_sizes[1][1]
        max_tests = self.max_tests if self.max_tests > 0 else max_samples_available

        pval_list, samples_used, df_list = [], [], []
        n = 0

        while n < max_samples_available - 1:
            valid_sources = [source for source, size in sample_sizes if size > n + 1]
            if len(valid_sources) < self.min_test_sample_size:
                n += 2
                continue

            df_tmp = self._construct_pair_df(
                grouped_data,
                covariate_vars,
                treatment_var,
                outcome_var,
                valid_sources,
                n,
            )
            df_list.append(df_tmp)
            n += 2

            if len(df_list) >= max_tests:
                break

        if not df_list:
            raise ValueError("No valid sample pairs available for testing.")

        cond_vars = (
            [f"{v}_i" for v in covariate_vars]
            + [f"{v}_j" for v in covariate_vars]
            + ["treatment_i"]
        )

        if self.cond_indep_test == "kcit_rbf":
            test_func = kcit_rbf
        elif self.cond_indep_test == "fisherz":
            test_func = fisherz
        else:
            raise ValueError(f"Unsupported cond_indep_test: {self.cond_indep_test}")

        for df in df_list:
            pval = test_func(
                df["treatment_j"].values.reshape(-1, 1),
                df["outcome_i"].values.reshape(-1, 1),
                df[cond_vars].values,
            )

            if np.isnan(pval):
                print("Warning: Computed NaN p-value.")
            else:
                pval_list.append(pval)
                samples_used.append(len(df))

        if not pval_list:
            raise RuntimeError("No valid p-values were computed.")

        weights = (
            [np.sqrt(n) for n in samples_used]
            if self.method_pval_combination == "stouffer"
            else None
        )
        _, combined_pval = combine_pvalues(
            pval_list, method=self.method_pval_combination, weights=weights
        )

        return combined_pval

    def _construct_pair_df(
        self,
        data: Dict[str, pd.DataFrame],
        covariates: List[str],
        treatment_var: str,
        outcome_var: str,
        sources: List[str],
        index: int,
    ) -> pd.DataFrame:
        """
        Construct a pairwise sample DataFrame for statistical testing.

        Args:
            data (Dict[str, pd.DataFrame]): Environment data.
            sources (List[str]): List of source keys to sample from.
            covariates (List[str]): List of observed covariates.
            index (int): Starting index for sampling.

        Returns:
            pd.DataFrame: Constructed sample DataFrame.
        """
        get = lambda source, col: data[source][col].values

        df_data = {
            "treatment_i": np.array(
                [get(source, treatment_var)[index] for source in sources]
            ),
            "outcome_i": np.array(
                [get(source, outcome_var)[index] for source in sources]
            ),
            "treatment_j": np.array(
                [get(source, treatment_var)[index + 1] for source in sources]
            ),
            "outcome_j": np.array(
                [get(source, outcome_var)[index + 1] for source in sources]
            ),
        }

        for var in covariates:
            df_data[f"{var}_i"] = np.array(
                [get(source, var)[index] for source in sources]
            )
            df_data[f"{var}_j"] = np.array(
                [get(source, var)[index + 1] for source in sources]
            )

        return pd.DataFrame(data=df_data)



import numpy as np
import pandas as pd
from typing import List, Dict
from jax import random, vmap, config
import jax.numpy as jnp

config.update("jax_enable_x64", True)

from causal_falsify.abstract import FalsificationAlgorithm
from causal_falsify.utils.mint import (
    create_polynomial_representation,
    bootstrap_model_fitting_jax,
    fit_outcome_model_jax,
    fit_treatment_model_jax,
    permutation_independence_test,
    bootstrapped_permutation_independence_test,
)


class MINT(FalsificationAlgorithm):
    def __init__(
        self,
        feature_representation: str = "linear",
        feature_representation_params: dict = {},
        binary_treatment: bool = False,
        min_samples_per_env: int = 25,
        independence_test_args: dict = {},
        n_bootstraps: int = 1000,
    ) -> None:
        """

        Mechanism INdependent Test (MINT) algorithm from
        "Falsification of Unconfoundedness by Testing Independence of Causal Mechanisms"
        Karlsson and Krijthe, ICML 2025
        (https://arxiv.org/abs/2502.06231)

        Joint test for whether we have independence between causal mechanisms and unconfoundedness across sources.
        A rejection will falsify both conditions jointly.

        Args:
            feature_representation (str): Feature representation to use ("linear" or "poly").
            feature_representation_params (dict): Parameters for the feature representation.
            binary_treatment (bool): Whether the treatment is binary (currently not implemented).
            min_samples_per_env (int): Minimum number of samples required per environment.
            independence_test_args (dict): Arguments for the independence test.
            n_bootstraps (int): Number of bootstrap iterations. If None, no bootstrapping is used.
        """
        self.feature_representation = feature_representation
        self.feature_representation_params = feature_representation_params
        self.binary_treatment = binary_treatment
        self.min_samples_per_env = min_samples_per_env
        self.independence_test_args = independence_test_args
        self.n_bootstraps = n_bootstraps

        # Store last diagnostics for access via get_diagnostics()
        self._last_model_fit_diagnostics = None

    def test(
        self,
        data: pd.DataFrame,
        covariate_vars: List[str],
        treatment_var: str,
        outcome_var: str,
        source_var: str,
    ) -> float:
        """
        Perform falsification test for joint test of unconfoundedness and independence of causal mechanisms.

        Args:
            data (dict): Dictionary mapping environment name to pd.DataFrame.
            covariate_vars (list): List of covariate column names.
            treatment_var (str): Name of treatment column.
            outcome_var (str): Name of outcome column.
            source_var (str): Name of source/environment column.

        Returns:
            p-value
        """

        # Validate required columns
        required_cols = set(covariate_vars + [treatment_var, outcome_var, source_var])
        missing = required_cols.difference(data.columns)
        if missing:
            raise ValueError(f"Missing columns in data: {missing}")

        # Extract arrays for the test
        outcome = data[[outcome_var]].values  # shape (n_samples, 1)
        treatment = data[[treatment_var]].values  # shape (n_samples, 1)
        source = data[[source_var]].values  # shape (n_samples, 1)
        covariates = data[covariate_vars].values  # shape (n_samples, n_covariates)

        n_environments = len(np.unique(source))
        coef_outcome_mech, coef_treatment_mech = [], []
        resampled_coef_outcome_mech, resampled_coef_treatment_mech = [], []

        model_fit_diagnostics = {
            "source_label": [],
            "outcome_model_mse": [],
            "treatment_model_mse": [],
        }

        # Prepare global feature mappings
        phi_outcome = self.get_feature_representation()
        phi_outcome_treatment = self.get_feature_representation()

        for source_label in np.unique(source):

            source_index = (source == source_label).squeeze()
            covariates_source = jnp.array(covariates[source_index, :])
            treatment_source = jnp.array(treatment[source_index, :])
            outcome_source = jnp.array(outcome[source_index, :])

            if covariates_source.shape[0] < self.min_samples_per_env:
                n_environments -= 1
                continue

            def add_intercept(term):
                return jnp.hstack([term, jnp.ones((term.shape[0], 1))])

            tf_outcome_source = add_intercept(phi_outcome(covariates_source))
            tf_outcome_treatment_source = add_intercept(
                phi_outcome_treatment(
                    jnp.concatenate([covariates_source, treatment_source], axis=1)
                )
            )

            params_outcome_mech, outcome_model_mse = fit_outcome_model_jax(
                tf_outcome_treatment_source, outcome_source
            )
            params_treatment_mech, treatment_model_mse = fit_treatment_model_jax(
                tf_outcome_source, treatment_source
            )

            coef_outcome_mech.append(params_outcome_mech)
            coef_treatment_mech.append(params_treatment_mech)

            model_fit_diagnostics["source_label"].append(source_label)
            model_fit_diagnostics["outcome_model_mse"].append(outcome_model_mse)
            model_fit_diagnostics["treatment_model_mse"].append(treatment_model_mse)

            if self.n_bootstraps:
                keys = random.split(random.PRNGKey(0), self.n_bootstraps)
                resampled_params = vmap(
                    bootstrap_model_fitting_jax,
                    in_axes=(None, None, None, None, 0),
                )(
                    outcome_source,
                    treatment_source,
                    tf_outcome_source,
                    tf_outcome_treatment_source,
                    keys,
                )
                resampled_coef_outcome_mech.append(resampled_params[0])
                resampled_coef_treatment_mech.append(resampled_params[1])

        coef_outcome_mech = np.array(jnp.vstack(coef_outcome_mech))
        coef_treatment_mech = np.array(jnp.vstack(coef_treatment_mech))

        if self.n_bootstraps:
            resampled_coef_outcome_mech = np.array(
                jnp.hstack(resampled_coef_outcome_mech)
            )
            resampled_coef_treatment_mech = np.array(
                jnp.hstack(resampled_coef_treatment_mech)
            )
            pval = self.run_bootstrapped_independence_test(
                coef_outcome_mech,
                coef_treatment_mech,
                resampled_coef_outcome_mech,
                resampled_coef_treatment_mech,
            )
        else:
            pval = self.run_independence_test(coef_outcome_mech, coef_treatment_mech)

        # save diagnostics from this run
        self._last_model_fit_diagnostics = model_fit_diagnostics

        return pval

    def get_diagnostics(self) -> Dict:
        """
        TODO Returns diagnostics from the last test run.

        Returns:
            dict: model_fit_diagnostics
        """
        return self._last_model_fit_diagnostics

    def get_feature_representation(self):
        if self.feature_representation == "linear":
            return lambda x: x
        elif self.feature_representation == "poly":
            return lambda x: create_polynomial_representation(
                x, **self.feature_representation_params
            )
        else:
            raise ValueError(
                f"Invalid feature representation: {self.feature_representation}"
            )

    def run_independence_test(self, data_x, data_y):
        return permutation_independence_test(data_x=data_x, data_y=data_y)

    def run_bootstrapped_independence_test(
        self, data_x, data_y, resampled_data_x, resampled_data_y
    ):
        return bootstrapped_permutation_independence_test(
            data_x=data_x,
            data_y=data_y,
            resampled_data_x=resampled_data_x,
            resampled_data_y=resampled_data_y,
        )



import numpy as np
import pandas as pd
import pytest

from causal_falsify.utils.simulate_data import (
    create_polynomial_representation,
    simulate_data,
)


# --------------------------
# Tests for create_polynomial_representation
# --------------------------


def test_create_polynomial_correct_shape():
    X = np.array([[1, 2], [3, 4]])
    degree = 3
    X_poly = create_polynomial_representation(X, degree)
    assert X_poly.shape == (2, 6), "Incorrect shape of polynomial features"


def test_create_polynomial_degree_1_equals_input():
    X = np.random.rand(10, 3)
    X_poly = create_polynomial_representation(X, degree=1)
    np.testing.assert_almost_equal(X, X_poly)


def test_create_polynomial_raises_on_non_array():
    with pytest.raises(TypeError):
        create_polynomial_representation([[1, 2], [3, 4]], degree=2)


def test_create_polynomial_raises_on_wrong_dim():
    with pytest.raises(ValueError):
        create_polynomial_representation(np.array([1, 2, 3]), degree=2)


def test_create_polynomial_raises_on_bad_degree():
    X = np.random.rand(5, 2)
    with pytest.raises(ValueError):
        create_polynomial_representation(X, 0)

    with pytest.raises(TypeError):
        create_polynomial_representation(X, "2")


def test_create_polynomial_raises_on_empty_input():
    with pytest.raises(ValueError):
        create_polynomial_representation(np.empty((0, 2)), 2)

    with pytest.raises(ValueError):
        create_polynomial_representation(np.empty((3, 0)), 2)


# --------------------------
# Tests for simulate_data
# --------------------------


def test_simulate_data_shape():
    df = simulate_data(n_samples=10, n_envs=3, n_observed_confounders=2, seed=42)
    expected_rows = 10 * 3
    expected_cols = 2 + 2 + 1  # A, Y, X_0, X_1, S
    assert df.shape == (expected_rows, expected_cols)
    assert set(["A", "Y", "X_0", "X_1", "S"]).issubset(df.columns)


def test_simulate_data_reproducibility():
    df1 = simulate_data(n_samples=5, seed=123)
    df2 = simulate_data(n_samples=5, seed=123)
    pd.testing.assert_frame_equal(df1, df2)


def test_simulate_data_env_column():
    df = simulate_data(n_samples=2, n_envs=4)
    assert df["S"].nunique() == 4
    assert set(df["S"].unique()) == {0, 1, 2, 3}


def test_simulate_data_raises_on_negative_samples():
    with pytest.raises(ValueError):
        simulate_data(n_samples=-5)



import numpy as np
import pytest
import jax.numpy as jnp
from jax import random

from causal_falsify.utils.mint import (
    create_polynomial_representation,
    compute_offdiag_block_frobnorm,
    permutation_independence_test,
    bootstrapped_permutation_independence_test,
    validate_matrix,
    fit_logistic_regression,
    fit_linear_regression,
    cross_val_mse,
    fit_outcome_model_jax,
    fit_treatment_model_jax,
    bootstrap_model_fitting_jax,
    resample_until_enough_unique,
)


def test_create_polynomial_representation_basic():
    X = np.array([[1, 2], [3, 4]])
    # degree must be > 1, else error
    with pytest.raises(ValueError):
        create_polynomial_representation(X, degree=1)

    # basic polynomial expansion without sklearn
    poly = create_polynomial_representation(X, degree=2, use_sklearn=False)
    assert poly.shape == (2, 4)  # 2 features * degree 2
    # with sklearn interaction_only
    poly2 = create_polynomial_representation(
        X, degree=2, use_sklearn=True, interaction_only=True
    )
    assert poly2.shape[0] == X.shape[0]


def test_validate_matrix_good_and_bad():
    good = np.array([[1.0, 2.0], [3.0, 4.0]])
    validate_matrix(good)  # should not raise

    with pytest.raises(AssertionError):
        validate_matrix(np.array([1, 2, 3]))  # not 2D

    with pytest.raises(AssertionError):
        validate_matrix(np.array([[np.nan, 1], [2, 3]]))

    with pytest.raises(AssertionError):
        validate_matrix(np.array([[np.inf, 1], [2, 3]]))


def test_compute_offdiag_block_frobnorm_shape_and_value():
    np.random.seed(0)
    x = np.random.randn(5, 3)
    y = np.random.randn(5, 2)
    norm = compute_offdiag_block_frobnorm(x, y)
    assert isinstance(norm, float)
    assert norm >= 0


def test_permutation_independence_test_basic():
    np.random.seed(0)
    x = np.random.randn(10, 2)
    y = np.random.randn(10, 2)
    p_val = permutation_independence_test(
        x, y, n_bootstraps=10, random_state=np.random.RandomState(0)
    )
    assert 0 <= p_val <= 1


def test_bootstrapped_permutation_independence_test_basic():
    np.random.seed(0)
    x = np.random.randn(10, 2)
    y = np.random.randn(10, 2)
    resampled_x = np.random.randn(3, 10, 2)
    resampled_y = np.random.randn(3, 10, 2)
    p_val = bootstrapped_permutation_independence_test(
        x, y, resampled_x, resampled_y, random_state=np.random.RandomState(0)
    )
    assert 0 <= p_val <= 1


def test_fit_logistic_regression_basic():
    np.random.seed(0)
    X = jnp.array([[1, 2], [3, 4], [5, 6], [7, 8]])
    T = jnp.array([1, -1, 1, -1])
    params = fit_logistic_regression(X, T, alpha=0.01)
    assert params.shape == (X.shape[1],)


def test_fit_linear_regression_basic():
    np.random.seed(0)
    X = jnp.array([[1, 2], [3, 4], [5, 6], [7, 8]])
    Y = jnp.array([1, 2, 3, 4])
    params = fit_linear_regression(X, Y)
    assert params.shape == (X.shape[1],)


def test_cross_val_mse_basic():
    np.random.seed(0)
    X = jnp.array([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]])
    Y = jnp.array([1, 2, 3, 4, 5])
    mse = cross_val_mse(X, Y, fit_linear_regression, num_folds=5)
    assert mse >= 0


def test_fit_outcome_and_treatment_model_jax():
    np.random.seed(0)
    X = jnp.array(np.random.randn(10, 3))
    T = jnp.array(np.random.randn(10))
    Y = jnp.array(np.random.randn(10))

    params_outcome, mse_outcome = fit_outcome_model_jax(X, Y)
    params_treatment, mse_treatment = fit_treatment_model_jax(X, T)

    assert params_outcome.shape == (X.shape[1],)
    assert mse_outcome >= 0
    assert params_treatment.shape == (X.shape[1],)
    assert mse_treatment >= 0


def test_resample_until_enough_unique():
    key = random.PRNGKey(0)
    indices = resample_until_enough_unique(key, n_resamples=10, min_sample_size=5)
    assert len(indices) == 10
    assert len(set(indices.tolist())) >= 5


def test_bootstrap_model_fitting_jax_basic():
    key = random.PRNGKey(0)
    np.random.seed(0)
    Y = jnp.array(np.random.randn(15))
    T = jnp.array(np.random.randn(15))
    tf_X = jnp.array(np.random.randn(15, 3))
    tf_XT = jnp.array(np.random.randn(15, 3))
    params_outcome, params_treatment = bootstrap_model_fitting_jax(
        Y, T, tf_X, tf_XT, key
    )
    assert params_outcome.shape == (tf_XT.shape[1],)
    assert params_treatment.shape == (tf_X.shape[1],)



import numpy as np
import pandas as pd


def create_polynomial_representation(X, degree):
    """
    Create polynomial features for input array X up to a given degree.

    Parameters:
    - X: np.ndarray, shape (n_samples, n_features)
    - degree: int, degree of polynomial features

    Returns:
    - np.ndarray, shape (n_samples, n_features * degree)
    """

    # Validate input type
    if not isinstance(X, np.ndarray):
        raise TypeError("X must be a NumPy array.")

    if X.ndim != 2:
        raise ValueError("X must be a 2D array of shape (n_samples, n_features).")

    if not isinstance(degree, int):
        raise TypeError("degree must be an integer.")

    if degree < 1:
        raise ValueError("degree must be a positive integer (>= 1).")

    n_samples, n_features = X.shape

    if n_samples == 0 or n_features == 0:
        raise ValueError("X must have at least one sample and one feature.")

    # Create polynomial features
    poly_features = []
    for feature_idx in range(n_features):
        feature = X[:, feature_idx]
        poly_feature = np.column_stack([feature**d for d in range(1, degree + 1)])
        poly_features.append(poly_feature)

    X_poly = np.hstack(poly_features)
    return X_poly


def simulate_data(
    n_samples: int,
    degree: int = 1,
    conf_strength: float = 1.0,
    transportability_violation: float = 0.0,
    n_envs: int = 50,
    n_observed_confounders: int = 5,
    seed: int = None,
) -> pd.DataFrame:

    rng = np.random.RandomState(seed)
    covar_list = [f"X_{i}" for i in range(n_observed_confounders)]

    x_transform = (
        (lambda x: x)
        if degree == 1
        else (lambda x: create_polynomial_representation(x, degree))
    )

    test_vector = rng.multivariate_normal(
        np.zeros(n_observed_confounders), np.eye(n_observed_confounders), size=(1)
    )
    feature_dim = x_transform(test_vector).shape[1]

    x_to_a_coef = rng.choice([-1.0, 1.0], size=(feature_dim, n_envs))
    x_to_y_coef = np.ones((feature_dim, n_envs), dtype=float)
    a_to_y_effect = np.ones((1, n_envs), dtype=float)

    # Add random noise to coefficients across environments (if )
    # sigma = 0.0
    # x_to_a_coef += rng.normal(0, sigma, size=(feature_dim, n_envs))
    # x_to_y_coef += rng.normal(0, sigma, size=(feature_dim, n_envs))
    # _to_y_effect += rng.normal(0, sigma, size=(1, n_envs))

    intercept_a = rng.normal(0, 1.0, size=(1, n_envs))
    intercept_y = rng.normal(0, 1.0, size=(1, n_envs))

    mu_X = rng.normal(0, 1.0, size=(n_envs, n_observed_confounders))
    mu_U = rng.normal(0, 1.0, size=(n_envs, 1))
    X_cov = np.full((n_observed_confounders, n_observed_confounders), 0.1)
    np.fill_diagonal(X_cov, 2)
    sigma_U = 2.0

    all_data = []

    for i in range(n_envs):
        X = rng.multivariate_normal(
            mu_X[i], X_cov / np.sqrt(n_observed_confounders), size=n_samples
        )
        X_repr = x_transform(X)
        U = rng.normal(mu_U[i], sigma_U, size=(n_samples, 1))

        treatment_confounding = 1.0 if conf_strength != 0.0 else 0.0
        A = (
            intercept_a[:, i]
            + (X_repr @ x_to_a_coef[:, i]).reshape(-1, 1)
            + treatment_confounding * np.sum(np.abs(U), axis=1).reshape(-1, 1)
            + rng.normal(0, 0.5, size=(n_samples, 1))  # Treatment noise
        )

        Y = (
            transportability_violation * intercept_y[:, i]
            + (X_repr @ x_to_y_coef[:, i]).reshape(-1, 1)
            + conf_strength * np.sum(np.abs(U), axis=1).reshape(-1, 1)
            + (a_to_y_effect[:, i] * A).reshape(-1, 1)
            + rng.normal(0, 0.5, size=(n_samples, 1))  # Outcome noise
        )

        df_env = pd.DataFrame(
            np.concatenate([A, Y, X], axis=1),
            columns=["A", "Y"] + covar_list,
        )
        df_env["S"] = i
        all_data.append(df_env)

    df_all = pd.concat(all_data, ignore_index=True)
    return df_all
